/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.rest;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.env.Environment;
import org.ow2.bonita.env.EnvironmentFactory;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.exception.BonitaInternalException;
import org.ow2.bonita.facade.interceptor.InterceptorUtil;
import org.ow2.bonita.facade.rest.wrapper.RESTCommand;
import org.ow2.bonita.facade.rest.wrapper.RESTMap;
import org.ow2.bonita.facade.rest.wrapper.RESTObject;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.identity.auth.DomainOwner;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

/**
 * 
 * @author Elias Ricken de Medeiros, Matthieu Chaffotte
 *
 */
public class RESTServerAPIInterceptor implements InvocationHandler {

  private static final Logger LOG = Logger.getLogger(RESTServerAPIInterceptor.class.getName());
  
  private final Object api; 
  private boolean envCreated = false;

  public RESTServerAPIInterceptor(final Object api) {
    this.api = api;    
  }

  public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {
    try {
      Class< ? >[] params = method.getParameterTypes();
			final Method m = this.api.getClass().getMethod(method.getName(), params);

			//get current ClassLoader
			ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
			Environment environment = null;
			try {
				environment = getEnvironment();
				ClassLoader localClassLoader = null;

				if ((m.getName().equals("setVariable") || m.getName().equals("setProcessInstanceVariable") || 
						m.getName().equals("setActivityInstanceVariable")) && args.length == 3
				    && args[2] instanceof RESTObject) {
				  ProcessDefinitionUUID processUUID = null;
          if (args[0] instanceof ActivityInstanceUUID) {
            ActivityInstanceUUID activityUUID = (ActivityInstanceUUID) args[0];
            ActivityInstance activity = EnvTool.getAllQueriers().getActivityInstance(activityUUID);
            processUUID = activity.getProcessDefinitionUUID();
          } else if (args[0] instanceof ProcessInstanceUUID){
            ProcessInstanceUUID instanceUUID = (ProcessInstanceUUID) args[0];
            ProcessInstance instance = EnvTool.getAllQueriers().getProcessInstance(instanceUUID);
            processUUID = instance.getProcessDefinitionUUID();
          }
          
          args[2] = getObjectFromRESTObject(((RESTObject) args[2]), processUUID);
          
				} else if (m.getName().equals("execute") && args.length >= 1
				    && args[0] instanceof RESTCommand<?>) {					
					// commomClassLoader
					if (args.length == 1) {
						localClassLoader = EnvTool.getClassDataLoader()
						    .getCommonClassLoader();
					} else if (args.length == 2
					    && args[1] instanceof ProcessDefinitionUUID) {
						// processClassLoader
						localClassLoader = EnvTool.getClassDataLoader()
						    .getProcessClassLoader((ProcessDefinitionUUID) args[1]);
					}
					Thread.currentThread().setContextClassLoader(localClassLoader);

					args[0] = ((RESTCommand<?>) args[0]).getCommand();

				} else if (m.getName().equals("instantiateProcess") && args.length == 3 &&
						args[1] instanceof RESTMap<?, ?>){					
						args[1] = ((RESTMap<?, ?>)args[1]).getActualMap();
				} else if (m.getName().equals("getModifiedJavaObject") && (args[0] instanceof ProcessDefinitionUUID)
				    && args.length == 4) {
				  ProcessDefinitionUUID processUUID = (ProcessDefinitionUUID) args[0];
          if (args[2] instanceof RESTObject) {
            args[2] = getObjectFromRESTObject((RESTObject) args[2], processUUID);
          }
          if (args[3] instanceof RESTObject) {
            args[3] = getObjectFromRESTObject((RESTObject) args[3], processUUID);
          }
        }
			} catch (Exception e) {
			  if (LOG.isLoggable(Level.SEVERE)) {
		      LOG.severe("Error on RESTServerInterceptor: " + Misc.getStackTraceFrom(e));
		    }
			} finally {
				if (this.envCreated) {
					environment.close();
				}
				Thread.currentThread().setContextClassLoader(baseClassLoader);
			}

			Object ret = m.invoke(this.api, args);
			//workaround : to be sure that GenericObjectProvider will be used
			//even if the object is instance of a primitive type 
			if (isAmbigousReturnType(m, ret)){
				ret = new RESTObject((Serializable)ret);
			}
			return ret;
   } catch (final Exception e) {
      if (e instanceof InvocationTargetException) {
        final Throwable invocationExceptionCause = e.getCause();
        if (invocationExceptionCause instanceof RemoteException) {
          final RemoteException remoteException = (RemoteException) invocationExceptionCause;
          final Throwable remoteCause = getRemoteCause(remoteException);
          InterceptorUtil.manageInvokeExceptionCause(method, remoteCause);
        } else {
          throw invocationExceptionCause;
        }
      }
      String message = ExceptionManager.getInstance().getFullMessage("baa_CAPII_1", e);
      throw new BonitaInternalException(message, e);
    }
  }

  private boolean isAmbigousReturnType(Method m, Object ret) {
  	if (ret == null || !ret.getClass().isArray()){
  		return false;
  	}

  	String methodName = m.getName();
  	//methods returning an Object or a generic unknowed type 
		return methodName.equals("execute") 
		|| methodName.equals("getVariable")
		|| methodName.equals("getActivityInstanceVariable")
		|| methodName.equals("getProcessInstanceVariable")
		|| methodName.equals("evaluateGroovyExpression");
	}
  
  private Object getObjectFromRESTObject (RESTObject restObject, ProcessDefinitionUUID processUUID) throws IOException, ClassNotFoundException{
    Object result = null;
    ClassLoader localClassLoader = null;
    try {
      result = restObject.getObject();
    } catch (ClassNotFoundException e) {
      // try with ProcessClassLoader
      localClassLoader = EnvTool.getClassDataLoader().getProcessClassLoader(processUUID);
      Thread.currentThread().setContextClassLoader(localClassLoader);

      try {
        result = restObject.getObject();
      } catch (ClassNotFoundException ex) {
        // try with CommonClassLoader
        localClassLoader = EnvTool.getClassDataLoader()
            .getCommonClassLoader();
        Thread.currentThread().setContextClassLoader(localClassLoader);

        result = restObject.getObject();
      }
    }
    
    return result;
  }

	private Throwable getRemoteCause(final RemoteException e) {
    Throwable t = e;
    while (t instanceof RemoteException) {
      t = t.getCause();
    }
    return t;
  }

  private Environment getEnvironment() throws Exception{
  	Environment environment = Environment.getCurrent();
		if (environment == null) {
			this.envCreated = true;
			EnvironmentFactory envFactory = GlobalEnvironmentFactory
			    .getEnvironmentFactory(DomainOwner.getDomain());
			environment = envFactory.openEnvironment();
		}
		return environment;
  }

}
